/*
 * Local Matters, Inc. Confidential
 * OCO Source Materials
 * (C) Local Matters, Inc. 2007. All Rights Reserved.
 */
package com.localmatters.flexiq.collector.producer;

import static com.localmatters.flexiq.collector.CollectorConstants.*;
import com.localmatters.flexiq.security.CipherAlgorithmEnum;
import com.localmatters.flexiq.security.CipherService;
import com.localmatters.flexiq.security.CipherServiceBase;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.security.GeneralSecurityException;
import java.security.Key;
import java.security.MessageDigest;
import java.util.zip.Deflater;
import javax.annotation.PostConstruct;
import javax.annotation.PreDestroy;
import javax.annotation.Resource;
import javax.crypto.Cipher;
import javax.crypto.spec.SecretKeySpec;
import javax.ejb.Stateless;
import javax.jms.BytesMessage;
import javax.jms.Connection;
import javax.jms.ConnectionFactory;
import javax.jms.DeliveryMode;
import javax.jms.Destination;
import javax.jms.JMSException;
import javax.jms.MessageProducer;
import javax.jms.Session;
import javax.jws.WebMethod;
import javax.jws.WebParam;
import javax.jws.WebService;
import javax.naming.InitialContext;
import javax.xml.XMLConstants;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.stream.StreamSource;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;
import javax.xml.validation.Validator;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathFactory;
import org.apache.commons.codec.binary.Hex;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

/**
 * The collector message producer implementation.
 *
 * @author Daniel Cavalcanti
 */
@Stateless()
@WebService()
public class ProducerWS {
    
    // **********************************************************************
    // Class Constants
    // **********************************************************************
    
    /**
     * The collector-messages schema relative to the application packaging root context.
     */
    public static final String COLLECTOR_MESSAGES_SCHEMA_LOCATION = "META-INF/collector-message.xsd";
    
    /**
     * The xpath expression to retrieve all collector-messages header elements.
     */
    public static final String HEADER_ELEMENTS_XPATH = "/collector-message/headers/header";
    
    // **********************************************************************
    // Instance variables
    // **********************************************************************
    
    // Logging facility *****************************************************
    
    private Log log = LogFactory.getLog(getClass());
    
    // JMS objects **********************************************************
    
    @Resource(mappedName="jms/CollectorConnectionFactory")
    private ConnectionFactory connectionFactory;
    
    @Resource(mappedName="jms/InitialDestination")
    private Destination initialDestination;
    
    private Connection connection;
    private Session session;
    private MessageProducer producer;
    
    // Injected resources ***************************************************
    
    @Resource(name="username")
    private String username;
    
    @Resource(name="password")
    private String password;
    
    @Resource(name="persist")
    private boolean persist = true;
    
    @Resource(name="validate")
    private boolean validate = false;
    
    @Resource(name="compress")
    private boolean compress = false;
    
    @Resource(name="encryption-algorithm")
    private String encryptionAlgorithm;
    
    @Resource(name="encrypt")
    private boolean encrypt = false;
    
    @Resource(name="encryption-key")
    private String encodedEncryptionKey;
    
    // Helper objects *******************************************************
    
    private Hex hexCodec = new Hex();
    private MessageDigest digester;
    
    private Schema schema;
    private Validator validator;
    
    private Deflater compressor;
    private CipherService cipherService;
    
    // **********************************************************************
    // Constructors
    // **********************************************************************
    
    /**
     * Creates a new instance of ProducerWS.
     */
    public ProducerWS() {
        if (log.isTraceEnabled()) log.trace("Creating collector producer bean.");
    }
    
    // **********************************************************************
    // Life-cycle methods
    // **********************************************************************
    
    /**
     * Initialize resources.
     */
    @PostConstruct()
    private void init() {
        
        // ******************************************************************
        // Log initial parameters
        // ******************************************************************
        
        if (log.isTraceEnabled()) {
            
            final String endl = System.getProperty("line.separator", "\n");
            StringBuilder buf = new StringBuilder();
            
            buf.append("Initializing collector producer bean.").append(endl);
            buf.append("The following producer parameters apply:").append(endl);
            buf.append("username.......: ").append(username).append(endl);
            buf.append("persist........: ").append(persist).append(endl);
            buf.append("validate.......: ").append(validate).append(endl);
            buf.append("compress.......: ").append(compress).append(endl);
            buf.append("encrypt........: ").append(encrypt).append(endl);
            buf.append("encryption-key.: ").append(encodedEncryptionKey).append(endl);
            
            log.debug(buf.toString());
            
        }
        
        // ******************************************************************
        // Establish JMS connectivity
        // ******************************************************************
        
        try {
            
            if (connectionFactory == null) {
                if (log.isTraceEnabled()) log.trace("Connection factory not set. Retrieving from server context.");
                InitialContext ic = new InitialContext();
                connectionFactory = (ConnectionFactory) ic.lookup("jms/CollectorConnectionFactory");
            }
            
            if (log.isTraceEnabled()) log.trace("Creating JMS connection.");
            connection = connectionFactory.createConnection(username, password);
            
            if (log.isTraceEnabled()) log.trace("Creating transacted session with JMS broker.");
            session = connection.createSession(true, Session.SESSION_TRANSACTED);
            
            if (log.isTraceEnabled()) log.trace("Creating message producer. Persistent? " + persist);
            producer = session.createProducer(initialDestination);
            producer.setDeliveryMode(persist ? DeliveryMode.PERSISTENT : DeliveryMode.NON_PERSISTENT);
            
            if (log.isTraceEnabled()) log.trace("Starting connection.");
            connection.start();
            
        } catch (Throwable ex) {
            log.error("Unexpected error.", ex);
        }
        
        // ******************************************************************
        // Message processors
        // ******************************************************************
        
        try {
            
            if (log.isTraceEnabled()) log.trace("Initializing message digester.");
            digester = MessageDigest.getInstance("MD5");
            
            if (log.isTraceEnabled()) log.trace("Initializing schema.");
            SchemaFactory schemaFactory = SchemaFactory.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);
            InputStream is = getClass().getClassLoader().getResourceAsStream("META-INF/collector-message.xsd");
            schema = schemaFactory.newSchema(new StreamSource(is));
            is.close();
            
            if (log.isTraceEnabled()) log.trace("Initializing schema validator.");
            validator = schema.newValidator();
            
            if (log.isTraceEnabled()) log.trace("Initializing message compressor.");
            compressor = new Deflater();
            
            if (log.isTraceEnabled()) log.trace("Loading cipher key.");
            byte[] decodedEncryptionKey = hexCodec.decodeHex(encodedEncryptionKey.toCharArray());
            
            if (ENCRYPTION_ALGORITHM_AES.equals(encryptionAlgorithm))
                cipherService = CipherServiceBase.getInstance(CipherAlgorithmEnum.AES, decodedEncryptionKey);
            else if (ENCRYPTION_ALGORITHM_BLOWFISH.equals(encryptionAlgorithm))
                cipherService = CipherServiceBase.getInstance(CipherAlgorithmEnum.Blowfish, decodedEncryptionKey);
            else if (ENCRYPTION_ALGORITHM_DES.equals(encryptionAlgorithm))
                cipherService = CipherServiceBase.getInstance(CipherAlgorithmEnum.DES, decodedEncryptionKey);
            else if (ENCRYPTION_ALGORITHM_TRIPLE_DES.equals(encryptionAlgorithm))
                cipherService = CipherServiceBase.getInstance(CipherAlgorithmEnum.TRIPLE_DES, decodedEncryptionKey);
            else
                throw new IllegalArgumentException("Invalid encryption algorithm: " + encryptionAlgorithm);
            
        } catch (Throwable ex) {
            log.error("Unexpected error.", ex);
        }
        
        if (log.isTraceEnabled()) log.trace("Collector producer bean started.");
        
    }
    
    /**
     * Releases resources.
     */
    @PreDestroy()
    private void release() {
        
        // ******************************************************************
        // Release connectivity
        // ******************************************************************
        
        try {
            
            if (connection != null) {
                if (log.isTraceEnabled()) log.trace("Releasing JMS connectivity.");
                connection.close();
            }
            
            if (log.isTraceEnabled()) log.trace("Collector producer bean JMS connectivity resources released.");
            
        } catch (JMSException ex) {
            log.error("Failed to release JMS connectivity.", ex);
        } catch (Throwable ex) {
            log.error("Unexpected error.", ex);
        }
        
    }
    
    // **********************************************************************
    // ProducerService methods
    // **********************************************************************
    
    /**
     * Sends a set of collectorMessages into the system.
     *
     * @param collectorMessages The list of collector messages to send into the system.
     * @param username The queue username.
     * @param password The queue password.
     * @throws com.localmatters.flexiq.collector.producer.incoming.ProducerException If there is an exception/error.
     */
    @WebMethod()
    public void send(@WebParam(name="msgs") byte[][] collectorMessages, @WebParam(name="username") String username, @WebParam(name="password") String password)
    throws ProducerException {
        
        // ******************************************************************
        // Validate parameters
        // ******************************************************************
        
        if (log.isTraceEnabled()) log.trace("Validating web-service parameters.");
        
        if (username == null)
            throw new ProducerException(new NullPointerException("Username cannot be null."));
        if (password == null)
            throw new ProducerException(new NullPointerException("Password cannot be null."));
        if (!username.equals(this.username) || !password.equals(this.password))
            throw new ProducerException(new IllegalArgumentException("Invalid user credentials."));
        if (collectorMessages == null)
            throw new ProducerException(new NullPointerException("Messages cannot be null."));
        for (byte [] message : collectorMessages)
            if (message == null)
                throw new ProducerException(new NullPointerException("Message cannot be null."));
        
        // ******************************************************************
        // Send collector messages
        // ******************************************************************
        
        try {
            
            if (log.isTraceEnabled()) log.trace("Creating re-usable objects: stream-source and jaxb-unmarshaller.");
            
            // a reusable StreamSource object
            StreamSource ss = new StreamSource();
            
            // a reusable Document Builder Factory
            DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory.newInstance();
            
            if (log.isTraceEnabled()) log.trace("Processing messages: " + collectorMessages.length);
            for (byte[] collectorMessage : collectorMessages) {
                
                if (log.isTraceEnabled()) log.trace("Processing message:\n" + new String(collectorMessage));
                
                ByteArrayInputStream bais = new ByteArrayInputStream(collectorMessage);
                ss.setInputStream(bais);
                
                // **************************************************************
                // Validate collector message
                // **************************************************************
                
                if (validate) {
                    if (log.isTraceEnabled()) log.trace("Validating message.");
                    validator.validate(ss);
                    validator.reset();
                    if (log.isTraceEnabled()) log.trace("Message validated.");
                }
                
                // reset stream so it can be re-used
                bais.reset();
                
                // **************************************************************
                // Create JMS message
                // **************************************************************
                
                if (log.isTraceEnabled()) log.trace("Creating JMS message.");
                BytesMessage jmsMessage = session.createBytesMessage();
                
                if (log.isTraceEnabled()) log.trace("Setting collector-message headers into JMS message.");
                Document dom = documentBuilderFactory.newDocumentBuilder().parse(bais);
                XPath xpath = XPathFactory.newInstance().newXPath();
                NodeList headerElements = (NodeList) xpath.evaluate("/collector-message/headers/header", dom, XPathConstants.NODESET);
                
                for (int i = 0 ; i < headerElements.getLength() ; i++) {
                    Element headerElement = (Element) headerElements.item(i);
                    jmsMessage.setStringProperty(headerElement.getAttribute("name"), headerElement.getAttribute("value"));
                }
                
                // **************************************************************
                // Digest
                // **************************************************************
                
                if (log.isTraceEnabled()) log.trace("Generating digital signature for collector message.");
                byte[] signature = digester.digest(collectorMessage);
                jmsMessage.setStringProperty(DIGITAL_SIGNATURE, new String(hexCodec.encodeHex(signature)));
                
                // **************************************************************
                // Compress message
                // **************************************************************
                
                jmsMessage.setBooleanProperty(COMPRESS_HEADER, compress);
                boolean performCompression = compress && (!jmsMessage.propertyExists(ASN_FORMAT) || !jmsMessage.getBooleanProperty(ASN_FORMAT));
                if (performCompression) {
                    if (log.isTraceEnabled()) log.trace("Compressing collector message.");
                    collectorMessage = compress(collectorMessage);
                }
                jmsMessage.setBooleanProperty(COMPRESSED_HEADER, performCompression);
                
                // **************************************************************
                // Encrypt message
                // **************************************************************
                
                jmsMessage.setBooleanProperty(ENCRYPT_HEADER, encrypt);
                boolean performEncryption = encrypt && (!jmsMessage.propertyExists(ASN_FORMAT) || !jmsMessage.getBooleanProperty(ASN_FORMAT));
                if (performEncryption) {
                    if (log.isTraceEnabled()) log.trace("Encrypting collector message.");
                    collectorMessage = encrypt(collectorMessage);
                    jmsMessage.setStringProperty(ENCRYPTION_ALGORITHM, encryptionAlgorithm);
                    jmsMessage.setStringProperty(ENCRYPTION_KEY, encodedEncryptionKey);
                }
                jmsMessage.setBooleanProperty(ENCRYPTED_HEADER, performEncryption);
                
                // **************************************************************
                // Conclude
                // **************************************************************
                
                if (log.isTraceEnabled()) log.trace("Writing collector message to JMS message.");
                jmsMessage.writeBytes(collectorMessage);
                
                if (log.isTraceEnabled()) log.trace("Sending JMS message.");
                producer.send(jmsMessage);
                if (log.isTraceEnabled()) log.trace("JMS message sent.");
                
            }
            
        } catch (Throwable ex) {
            log.error("Unexpected error.", ex);
            throw new ProducerException("Unexpected error.", ex);
        }
        
    }
    
    // **********************************************************************
    // Auxiliary methods
    // **********************************************************************
    
    /**
     * Compresses a collector message.
     *
     * @param collectorMessage The collector message.
     * @throws java.io.IOException If there is an exception/error.
     * @return The compressed collector message.
     */
    private byte[] compress(byte[] collectorMessage)
    throws IOException {
        
        compressor.reset();
        compressor.setInput(collectorMessage);
        compressor.finish();
        
        ByteArrayOutputStream baos = new ByteArrayOutputStream(collectorMessage.length);
        byte[] buf = new byte[1024];
        while (!compressor.finished()) {
            int count = compressor.deflate(buf);
            baos.write(buf, 0, count);
        }
        baos.close();
        
        if (log.isTraceEnabled()) log.trace("Message compressed.");
        
        return baos.toByteArray();
        
    }
    
    /**
     * Encrypts the collector message.
     *
     * @return The encrypted collector message.
     * @param collectorMessage The collector message.
     * @throws java.security.GeneralSecurityException If there is an error/exception.
     */
    private byte[] encrypt(byte[] collectorMessage)
    throws GeneralSecurityException {
        byte[] encrypted = cipherService.encrypt(collectorMessage);
        if (log.isTraceEnabled()) log.trace("Message encrypted.");
        return encrypted;
    }
    
}
