/*
 * Local Matters, Inc. Confidential
 * OCO Source Materials
 * (C) Local Matters, Inc. 2007. All Rights Reserved.
 */
package com.localmatters.flexiq.collector.router;

import static com.localmatters.flexiq.collector.CollectorConstants.*;
import java.io.ByteArrayInputStream;
import java.util.Enumeration;
import javax.annotation.PostConstruct;
import javax.annotation.PreDestroy;
import javax.annotation.Resource;
import javax.jms.BytesMessage;
import javax.jms.Connection;
import javax.jms.ConnectionFactory;
import javax.jms.DeliveryMode;
import javax.jms.Destination;
import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.MessageListener;
import javax.jms.MessageProducer;
import javax.jms.Session;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathFactory;
import org.apache.commons.codec.binary.Base64;
import org.apache.commons.codec.binary.Hex;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * Entity class RouterBean
 *
 * @author Daniel
 */
public class RouterBean
    implements MessageListener {
    
    /**
     * XPath expression to header that contains the XPath expression to the ASN
     * element that contains the legacy (V4) collector message.
     */
    public static final String HEADER_ASN_XPATH = "/collector-message/headers/header[@name='AsnXPath']/@value";
    
    // **********************************************************************
    // Instance variables
    // **********************************************************************
    
    // Logging facility *****************************************************
    
    /**
     * The class logger.
     */
    private Log log = LogFactory.getLog(getClass());
    
    // JMS objects **********************************************************
    
    @Resource(mappedName="jms/CollectorConnectionFactory")
    private ConnectionFactory connectionFactory;
    
    private Destination outgoingDestination;
    private Connection connection;
    private Session session;
    private MessageProducer producer;
    
    // Injected resources ***************************************************
    
    @Resource(name="username")
    private String username;
    
    @Resource(name="password")
    private String password;
    
    @Resource(name="outgoing-destination-jndi")
    private String outgoingDestinationJndi;
    
    @Resource(name="persist")
    private boolean persist;
    
    // Server context *******************************************************
    
    /**
     * The server initial context.
     */
    private InitialContext ic;
    
    // **********************************************************************
    // Constructors
    // **********************************************************************
    
    /**
     * Creates a new instance of RouterBean.
     */
    public RouterBean() {
        log.trace("creating router bean.");
    }
    
    // **********************************************************************
    // Life-cycle methods
    // **********************************************************************
    
    /**
     * Initialized the bean.
     */
    @PostConstruct()
    private void init() {
        
        // ******************************************************************
        // Log initial parameters
        // ******************************************************************
        
        if (log.isTraceEnabled()) log.trace("Initializing collector router bean.");
        
        // ******************************************************************
        // Establish JMS connectivity
        // ******************************************************************
        
        try {
            
            if (log.isTraceEnabled()) log.trace("Getting server initial context.");
            ic = new InitialContext();
            
            if (log.isTraceEnabled()) log.trace("Establishing connection to JMS broker.");
            connection = connectionFactory.createConnection(username, password);
            
            if (log.isTraceEnabled()) log.trace("Creating transacted session with JMS broker.");
            session = connection.createSession(true, session.SESSION_TRANSACTED);
            
            if (log.isTraceEnabled()) log.trace("Looking up for bound destination: " + outgoingDestinationJndi);
            outgoingDestination = getDestination(outgoingDestinationJndi);
            
            if (log.isTraceEnabled()) log.trace("Creating message producer for outgoing destination: " + outgoingDestinationJndi);
            producer = session.createProducer(outgoingDestination);
            producer.setDeliveryMode(persist ? DeliveryMode.PERSISTENT : DeliveryMode.NON_PERSISTENT);
            
            if (log.isTraceEnabled()) log.trace("Starting connection.");
            connection.start();
            
            if (log.isTraceEnabled()) log.trace("Collector router bean started successfully.");
            
        } catch (JMSException ex) {
            log.error("Failed to establish JMS connectivity.", ex);
        } catch (NamingException ex) {
            log.error("Failed to get server initial context.", ex);
        } catch (Throwable ex) {
            log.error("Unexpected error.", ex);
        }
        
    }
    
    /**
     * Releases bean resources.
     */
    @PreDestroy()
    private void release() {
        
        // ******************************************************************
        // Release connectivity
        // ******************************************************************
        
        try {
            
            if (connection != null) {
                if (log.isTraceEnabled()) log.trace("Releasing JMS connectivity.");
                connection.close();
            }
            
            if (log.isTraceEnabled()) log.trace("Collector router bean JMS connectivity resources released.");
            
        } catch (JMSException ex) {
            log.error("Failed to release JMS connectivity.", ex);
        } catch (Throwable ex) {
            log.error("Unexpected error.", ex);
        }
        
    }
    
    // **********************************************************************
    // MessageListener methods
    // **********************************************************************
    
    /**
     * Receives a message from the incoming destination and sends it to the
     * outgoing destination.
     *
     * @param message The message to route.
     */
    public void onMessage(Message message) {
        
        if (log.isTraceEnabled()) log.trace("Message received\n" + message);
        
        try {
            
            BytesMessage routed = session.createBytesMessage();
            
            if (log.isTraceEnabled()) log.trace("Copying message properties.");
            copyProperties(message, routed);
            
            if (log.isTraceEnabled()) log.trace("Copying message payload.");
            if (message.propertyExists(ASN_FORMAT) && message.getBooleanProperty(ASN_FORMAT))
                copyAsnContents((BytesMessage) message, routed);
            else
                copyContents((BytesMessage) message, routed);
            
            if (log.isTraceEnabled()) log.trace("Routing message to outgoing destination:\n" + outgoingDestination);
            producer.send(routed);
            if (log.isTraceEnabled()) log.trace("Message routed to outgoing destination:\n" + outgoingDestination);
            
        } catch (JMSException ex) {
            log.error("Unexpected error.");
            throw new RouterException("Unexpected error.", ex);
        } catch (Throwable ex) {
            log.error("Unexpected error.");
            throw new RouterException("Unexpected error.", ex);
        }
        
    }
    
    // **********************************************************************
    // Auxiliary methods
    // **********************************************************************
    
    /**
     * Returns a destination from the server context.
     *
     * @param jndiName The destination JNDI name.
     * @throws javax.naming.NamingException If there is an exception/error.
     * @return The destination bound to this name.
     */
    private Destination getDestination(String jndiName)
    throws NamingException {
        if (log.isTraceEnabled()) log.trace("Retrieving destination bound to " + jndiName);
        return (Destination) ic.lookup(jndiName);
    }
    
    /**
     * Copy the properties from source message to destination message.
     *
     * @param source The JMS source message.
     * @param destination The JMS destination message.
     * @throws javax.jms.JMSException If there is an exception/error.
     */
    private void copyProperties(Message source, Message destination)
    throws JMSException {
        for (Enumeration properties = source.getPropertyNames() ; properties.hasMoreElements() ; ) {
            String property = (String) properties.nextElement();
            destination.setStringProperty(property, source.getStringProperty(property));
        }
    }
    
    /**
     * Copy the message body from source message to destination message.
     *
     * @param source The JMS source message.
     * @param destination The JMS destination message.
     * @throws javax.jms.JMSException If there is an exception/error.
     */
    private void copyContents(BytesMessage source, BytesMessage destination)
    throws JMSException {
        byte[] buffer = new byte[(int) source.getBodyLength()];
        source.readBytes(buffer);
        destination.writeBytes(buffer);
    }
    
    /**
     * If source message content is ASN, the destination message if to be
     * constructed for legacy support.
     *
     * @param source The JMS source message.
     * @param destination The JMS destination message.
     * @throws java.lang.Exception If there is an exception/error.
     */
    private void copyAsnContents(BytesMessage source, BytesMessage destination)
    throws Exception {
        byte[] buffer = new byte[(int) source.getBodyLength()];
        source.readBytes(buffer);
        destination.writeBytes(retrieveLegacyMessage(buffer));
    }
    
    /**
     * Retrieves the legacy (V4) collector message embedded in the V5 collector
     * message DOM.
     *
     * @param buffer The V5 collector message DOM.
     * @return The legacy (v4) collector message.
     * @throws java.lang.Exception If there is an error/exception.
     */
    private byte[] retrieveLegacyMessage(byte[] buffer)
    throws Exception {
        
        // ******************************************************************
        // Parse collector message
        // ******************************************************************
        
        if (log.isTraceEnabled()) log.trace("Parsing collector message DOM.");
        
        ByteArrayInputStream is = new ByteArrayInputStream(buffer);
        DocumentBuilder db = DocumentBuilderFactory.newInstance().newDocumentBuilder();
        Document dom = db.parse(is);
        is.close();
        
        // ******************************************************************
        // Retrieve Base64 encoded legacy collector message
        // ******************************************************************
        
        if (log.isTraceEnabled()) log.trace("Retriving legacy collector message -- 64Base encoded.");
        
        XPath xpath = XPathFactory.newInstance().newXPath();
        Attr ASNPathValueAttr = (Attr) xpath.evaluate(HEADER_ASN_XPATH, dom, XPathConstants.NODE);
        Element node = (Element) xpath.evaluate(ASNPathValueAttr.getTextContent(), dom, XPathConstants.NODE);
        
        // ******************************************************************
        // Decode legacy collector message
        // ******************************************************************
        
        if (log.isTraceEnabled()) log.trace("Decoding legacy collector message.");
        
        Base64 base64Codec = new Base64();
        return base64Codec.decode(node.getTextContent().getBytes());
        
    }
    
}
